# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import override

import aiofiles

from dogcrud.core import context, rest, transformers
from dogcrud.core.pagination import PaginationStrategy
from dogcrud.core.resource_type import IDType, ResourceType

logger = logging.getLogger(__name__)


class StandardResourceType(ResourceType):
    """
    A Datadog Resource that operates like a standard CRUD object (although we
    only do RU operations).

    Attributes:
        get_to_put_transformer: used when the HTTP GET stores JSON that must be transformed before posting back to the PUT endpoint.
    """

    def __init__(
        self,
        rest_base_path: str,
        webpage_base_path: str,
        max_concurrency: int,
        pagination_strategy: PaginationStrategy,
        webpage_suffix: str = "",
        get_to_put_transformer: transformers.GetToPut = transformers.identity,
    ) -> None:
        self.rest_base_path = rest_base_path
        self.webpage_base_path = webpage_base_path
        self.concurrency_semaphore = asyncio.BoundedSemaphore(max_concurrency)
        self.pagination_strategy = pagination_strategy
        self.webpage_suffix = webpage_suffix
        self.get_to_put_transformer = get_to_put_transformer

    @override
    def rest_path(self, resource_id: IDType | None = None) -> str:
        match resource_id:
            case None:
                return self.rest_base_path
            case _:
                return f"{self.rest_base_path}/{resource_id}"

    @override
    def local_path(self, resource_id: IDType | None = None) -> Path:
        data_dir = context.config_context().data_dir
        match resource_id:
            case None:
                return data_dir / self.rest_path()
            case _:
                return data_dir / f"{self.rest_path(resource_id)}.json"

    @override
    async def get(self, resource_id: IDType) -> bytes:
        async with self.concurrency_semaphore:
            return await rest.get_json(f"api/{self.rest_path(resource_id)}")

    @override
    async def put(self, resource_id: IDType, data: bytes) -> None:
        async with self.concurrency_semaphore:
            await rest.put_json(f"api/{self.rest_path(resource_id)}", data)

    @override
    def transform_get_to_put(self, data: bytes) -> bytes:
        return self.get_to_put_transformer(data)

    @override
    async def list_ids(self) -> AsyncGenerator[IDType]:
        async for page in self.pagination_strategy.pages(f"api/{self.rest_path()}", self.concurrency_semaphore):
            for id_ in page.ids:
                yield id_

    @override
    async def read_local_json(self, resource_id: IDType) -> bytes:
        resource_dir = context.config_context().data_dir / self.rest_path(resource_id)
        filename = str(resource_dir / f"{resource_id}.json")

        async with aiofiles.open(filename, "rb") as file:
            return await file.read()

    @override
    def webpage_url(self, resource_id: IDType) -> str:
        return f"https://app.datadoghq.com/{self.webpage_base_path}/{resource_id}{self.webpage_suffix}"

    @override
    def resource_id(self, filename: str) -> IDType:
        return Path(filename).stem



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import override

import aiofiles
from pydantic import BaseModel

from dogcrud.core import context, rest
from dogcrud.core.pagination import CursorDataItemModel, CursorPagination
from dogcrud.core.resource_type import IDType, ResourceType

logger = logging.getLogger(__name__)


class MetricTagAggregationModel(BaseModel):
    space: str
    time: str


class MetricTagAttributesModel(BaseModel):
    aggregations: list[MetricTagAggregationModel] | None = None
    exclude_tags_mode: bool | None = None
    include_percentiles: bool | None = None
    tags: list[str] | None = None


class MetricTagDataModel(BaseModel):
    attributes: MetricTagAttributesModel | None = None
    id: str
    type: str


class MetricTagModel(BaseModel):
    data: MetricTagDataModel

class MetricResourceType(ResourceType):
    """
    A Datadog metric resource. This resource doesn't fit in with a StandardResourceType since
    - you have an idea of "active" metrics that the initial list is populated from from a time range
    - you get a list of metrics with a subset of metadata
    - there is no get ID list and get specific resource json like StandardResourceType
    """

    def __init__(
        self,
        max_concurrency: int,
    ) -> None:
        self.rest_base_path = "v2/metrics"
        self.concurrency_semaphore = asyncio.BoundedSemaphore(max_concurrency)
        self._id_to_item_index: dict[IDType, CursorDataItemModel] = {}

    @override
    def rest_path(self, resource_id: IDType | None = None) -> str:
        match resource_id:
            case None:
                return self.rest_base_path
            case _:
                return f"{self.rest_base_path}/{resource_id}"

    @override
    def local_path(self, resource_id: IDType | None = None) -> Path:
        data_dir = context.config_context().data_dir
        match resource_id:
            case None:
                return data_dir / self.rest_path()
            case _:
                return data_dir / f"{self.rest_path(resource_id)}.json"

    @override
    async def get(self, resource_id: IDType) -> bytes:
        # TODO: load _id_to_index_item if not yet set
        return self._id_to_item_index[resource_id].json().encode("utf-8")

    @override
    async def put(self, resource_id: IDType, data: bytes) -> None:
        url = f"api/v2/metrics/{resource_id}/tags"
        async with self.concurrency_semaphore:
            await rest.patch_json(url, data)

    @override
    def transform_get_to_put(self, data: bytes) -> bytes:
        metric_tag_data = MetricTagDataModel.model_validate_json(data)
        metric_tag = MetricTagModel(data=metric_tag_data)
        return metric_tag.model_dump_json(exclude_none=True)

    async def _list_ids(self, cursor_pagination: CursorPagination) -> AsyncGenerator[IDType]:
        async for page in cursor_pagination.pages(f"api/{self.rest_path()}", self.concurrency_semaphore):
            for item in page.data:
                self._id_to_item_index[item.id] = item
                yield item.id

    @override
    async def list_ids(self) -> AsyncGenerator[IDType]:
        # data[i].attributes is only set when filter[configured]=true is set,
        # so do 2 paginations—one with false and one with true—in order to get
        # the attributes.
        for configured in ("false", "true"):
            p = CursorPagination(query_params=f"filter[configured]={configured}")
            async for id_ in self._list_ids(p):
                yield id_

    @override
    async def read_local_json(self, resource_id: IDType) -> bytes:
        resource_dir = context.config_context().data_dir / self.rest_path(resource_id)
        filename = str(resource_dir / f"{resource_id}.json")

        async with aiofiles.open(filename, "rb") as file:
            return await file.read()

    @override
    def webpage_url(self, resource_id: IDType) -> str:
        return f"https://app.datadoghq.com/metric/summary?metric={resource_id}"

    @override
    def resource_id(self, filename: str) -> IDType:
        return Path(filename).stem



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging
from collections.abc import AsyncIterator, Sequence
from dataclasses import dataclass
from typing import Any, Protocol, assert_never, override

import orjson
from pydantic import BaseModel, ConfigDict

from dogcrud.core.resource_type import IDType
from dogcrud.core.rest import get_json

logger = logging.getLogger(__name__)


class Page(Protocol):
    def get_ids(self) -> Sequence[IDType]: ...
    def get_items(self) -> Sequence[Any]: ...
    def get_parsed_json(self) -> Any: ...


@dataclass(frozen=True)
class PageImpl:
    ids: Sequence[IDType]
    items: Sequence[Any]
    parsed_json: Any

    @override
    def get_ids(self) -> Sequence[IDType]:
        return self.ids

    @override
    def get_items(self) -> Sequence[Any]:
        return self.items

    @override
    def get_parsed_json(self) -> Any:
        return self.parsed_json


class PaginationStrategy(Protocol):
    def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]: ...


async def _get_page(url: str, items_key: str | None) -> Page:
    json = await get_json(url)
    parsed_json = orjson.loads(json)

    match (parsed_json, items_key):
        case list(), None:
            items = parsed_json
        case dict(), items_key if items_key is not None:
            items = parsed_json[items_key]
        case _, _:
            msg = f"Invalid combination of page response type {type(parsed_json)} and items key '{items_key}'"
            raise RuntimeError(msg)

    ids = [item["id"] for item in items]
    return PageImpl(ids=ids, items=items, parsed_json=parsed_json)


@dataclass(frozen=True)
class ItemOffsetPagination(PaginationStrategy):
    """
    A pagination strategy based on offsets.

    Attributes:
        offset_query_param: The name of the query parameter for the offset.
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    offset_query_param: str
    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        offset = 0
        seen_ids: set[IDType] = set()
        while True:
            async with concurrency_semaphore:
                page = await _get_page(f"{url}?{self.offset_query_param}={offset}", self.items_key)
            if not page.ids:
                return
            yield page
            offset += len(page.ids)
            new_ids = set(page.ids)
            if already_seen_ids := new_ids & seen_ids:
                logger.warning(
                    f"Duplicate IDs seen while paging. If transient okay to ignore cause it might be someone else mutating the list of items. already_seen_ids={already_seen_ids}. If not transient, this could be a misconfiguration of the pagination_strategy."
                )
            seen_ids |= new_ids


@dataclass(frozen=True)
class IDOffsetPagination(PaginationStrategy):
    """
    A pagination strategy based on last seen ID.

    Attributes:
        offset_query_param: The name of the query parameter for the offset.
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    offset_query_param: str
    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        id_offset: IDType = "0"
        offset = 0
        seen_ids: set[IDType] = set()
        while True:
            async with concurrency_semaphore:
                page = await _get_page(f"{url}?{self.offset_query_param}={id_offset}", self.items_key)
            if not page.ids:
                return
            yield page
            id_offset = page.ids[-1]
            offset += len(page.ids)
            new_ids = set(page.ids)
            if already_seen_ids := new_ids & seen_ids:
                logger.warning(
                    f"Duplicate IDs seen while paging. If transient okay to ignore cause it might be someone else mutating the list of items. already_seen_ids={already_seen_ids}. If not transient, this could be a misconfiguration of the pagination_strategy."
                )
            seen_ids |= new_ids


@dataclass(frozen=True)
class NoPagination(PaginationStrategy):
    """
    A pagination strategy that fetches everything in one go.

    Attributes:
        items_key: If the page response is an object, this is the key to the list of items. Use None if the response is already a list.
    """

    items_key: str | None = None

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[Page]:
        async with concurrency_semaphore:
            page = await _get_page(url, self.items_key)
        yield page


class CursorPaginationModel(BaseModel):
    next_cursor: str | None


class CursorMetadataModel(BaseModel):
    pagination: CursorPaginationModel


class CursorLinksModel(BaseModel):
    first: str | None
    last: str | None
    next: str | None
    prev: str | None
    self: str | None


class CursorDataItemModel(BaseModel):
    model_config = ConfigDict(extra='allow')

    id: str


class CursorPageModel(BaseModel):
    data: list[CursorDataItemModel]
    links: CursorLinksModel
    meta: CursorMetadataModel


@dataclass(frozen=True)
class CursorPagination(PaginationStrategy):
    """
    A pagination strategy based on last seen page cursor.
    """

    query_params: str = ""

    @override
    async def pages(self, url: str, concurrency_semaphore: asyncio.Semaphore) -> AsyncIterator[CursorPageModel]:
        next_url = f"{url}?{self.query_params}&page[size]=1000&page[cursor]="
        while True:
            async with concurrency_semaphore:
                json = await get_json(next_url)
                page = CursorPageModel.model_validate_json(json)

            yield page

            match page.links.next:
                case None:
                    return
                case str() as next:
                    next_url = next
                case never:
                    assert_never(never)



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import logging

import click

from dogcrud.core import data
from dogcrud.core.context import config_context
from dogcrud.core.resource_type import ResourceType
from dogcrud.core.resource_type_registry import resource_types

logger = logging.getLogger(__name__)


@click.group()
def save() -> None:
    """
    Save a datadog resources to the local file system.
    """


def create_save_command(rt: ResourceType):
    @save.command(
        name=rt.rest_path(),
        help=f"""
    Save {rt.rest_path()} resources.

    If RESOURCE_ID is all, save all {rt.rest_path()}
    resources; otherwise, save a single {rt.rest_path()} resource.
    """,
    )
    @click.argument("resource_id", type=str)
    def save_resource_command(resource_id: str):
        match resource_id:
            case _ if resource_id.lower() == "all":
                coro = save_all_resources_of_type(rt)
            case _:
                coro = save_resource(rt, resource_id)
        config_context().run_in_context(coro)


for rt in resource_types():
    create_save_command(rt)


@save.command(
    name="all",
    help=f"""
Save all datadog resources supported by this tool.

Includes: {", ".join(rt.rest_path() for rt in resource_types())}
""",
)
def save_all() -> None:
    """
    Save all datadog resources.
    """
    config_context().run_in_context(save_all_resources())


async def save_all_resources():
    async with asyncio.TaskGroup() as tg:
        for resource_type in resource_types():
            tg.create_task(save_all_resources_of_type(resource_type))


async def save_all_resources_of_type(resource_type: ResourceType) -> None:
    prefix = f"save all {resource_type.rest_path()}"
    logger.info(f"{prefix}: Starting")

    num_saved = 0

    async with asyncio.TaskGroup() as tg:
        async for resource_id in resource_type.list_ids():
            tg.create_task(save_resource(resource_type, str(resource_id)))
            num_saved += 1

    logger.info(f"{prefix} Saved {num_saved} items.")


async def save_resource(resource_type: ResourceType, resource_id: str) -> None:
    json = await resource_type.get(resource_id)
    resource_type.local_path().mkdir(exist_ok=True, parents=True)
    filename = resource_type.local_path(resource_id)
    await data.write_formatted_json(json, str(filename))
    logger.info(f"Saved {filename}")



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import logging
import pathlib
import resource

import click

from dogcrud.cli.open import open_in_browser
from dogcrud.cli.restore import restore
from dogcrud.cli.save import save
from dogcrud.core import context
from dogcrud.core.logging import setup_logger


@click.group()
@click.version_option()
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(list(logging.getLevelNamesMapping().keys()), case_sensitive=False),
)
@click.option(
    "--dd-api-key",
    type=str,
    envvar="DD_API_KEY",
    show_envvar=True,
    required=True,
    help="Datadog API key.",
)
@click.option(
    "--dd-app-key",
    type=str,
    envvar="DD_APP_KEY",
    show_envvar=True,
    required=True,
    help="Datadog application key.",
)
@click.option(
    "--max-concurrent-requests",
    type=int,
    default=100,
    help="Max concurrent requests to Datadog.",
)
@click.option(
    "--data-dir",
    type=click.Path(exists=True, file_okay=False, resolve_path=True, path_type=pathlib.Path),
    default="saved",
)
@click.option(
    "--min-open-files-limit",
    type=int,
    default=4096,
    help="Try to raise the ulimit if it's below this number. This tool opens lots of concurrent file handles.",
)
@click.pass_context
def cli(
    ctx: click.Context,
    log_level: str,
    dd_api_key: str,
    dd_app_key: str,
    max_concurrent_requests: int,
    data_dir: pathlib.Path,
    min_open_files_limit: int,
):
    """
    Utility for working with Datadog CRUD resources.

    For a full list of CRUD resources, see https://docs.datadoghq.com/api/latest/.

    Note, only a subset of those resources are currently supported by this tool.
    """
    root_logger = logging.getLogger()
    setup_logger(root_logger, log_level)

    # This tool opens a lot of file handles concurrently, so need to up limits.
    soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(
        resource.RLIMIT_NOFILE,
        (max(soft, min_open_files_limit), max(hard, min_open_files_limit)),
    )

    ctx.with_resource(
        context.set_config_context(
            context.ConfigContext(
                dd_api_key=dd_api_key,
                dd_app_key=dd_app_key,
                max_concurrent_requests=max_concurrent_requests,
                data_dir=data_dir,
            )
        )
    )


cli.add_command(save)
cli.add_command(restore)
cli.add_command(open_in_browser)

cli.main()



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import logging
import time


class ElapsedTimeFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(fmt="%(elapsed)s [%(levelname)s] %(message)s")
        self.now = time.monotonic
        self.t0 = self.now()

    def format(self, record):
        seconds_since_t0 = self.now() - self.t0
        record.elapsed = f"{seconds_since_t0:.2f}s"
        return super().format(record)


def setup_logger(logger: logging.Logger, level_name: str):
    handler = logging.StreamHandler()
    formatter = ElapsedTimeFormatter()
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.addHandler(handler)
    logger.setLevel(logging.getLevelNamesMapping()[level_name])
    return logger



# SPDX-FileCopyrightText: 2025-present Doug Richardson <git@rekt.email>
# SPDX-License-Identifier: MIT

import asyncio
import contextlib
import contextvars
import pathlib
from collections.abc import Coroutine, Generator
from dataclasses import dataclass

import aiohttp


class ContextAlreadySetError(Exception):
    def __init__(self, context_var_name: str):
        super().__init__(f"Context {context_var_name} already set.")


@dataclass(frozen=True)
class ConfigContext:
    """
    ConfigContext is set from the top level to setup common context variables
    in the async context. It is used to pass global options to sub-commands
    (like save and restore).

    Since the majority of the sub-command features use asyncio, but click is not
    asyncio compatible, ConfigContext is also used to startup the async sub-commands
    with an async context setup correctly.
    """

    dd_api_key: str
    dd_app_key: str
    max_concurrent_requests: int
    data_dir: pathlib.Path

    def run_in_context(self, main: Coroutine) -> None:
        asyncio.run(self.in_context(main))

    async def in_context(self, main: Coroutine) -> None:
        """
        Run coro with an AsyncRunContext configured.

        Access the context using async_run_context().
        """
        headers = {
            "DD-API-KEY": self.dd_api_key,
            "DD-APPLICATION-KEY": self.dd_app_key,
        }

        async with aiohttp.ClientSession(headers=headers) as datadog_session:
            token = async_run_context_var.set(
                AsyncRunContext(
                    datadog_session=datadog_session,
                    concurrent_requests_semaphore=asyncio.BoundedSemaphore(self.max_concurrent_requests),
                )
            )
            try:
                if token.old_value != contextvars.Token.MISSING:
                    raise ContextAlreadySetError(async_run_context_var.name)
                await main
            finally:
                async_run_context_var.reset(token)


config_context_var = contextvars.ContextVar[ConfigContext]("config_context")


def config_context() -> ConfigContext:
    return config_context_var.get()


@contextlib.contextmanager
def set_config_context(config_context: ConfigContext) -> Generator[None]:
    """
    Used by the top level click command to install the ConfigContext.
    """
    token = config_context_var.set(config_context)
    try:
        if token.old_value != contextvars.Token.MISSING:
            raise ContextAlreadySetError(config_context_var.name)
        yield
    finally:
        config_context_var.reset(token)


@dataclass
class AsyncRunContext:
    """
    AsyncRunContext is availble inside ConfigContext.run_in_context().


    Asyncio click sub-commands get this using async_run_context().
    """

    datadog_session: aiohttp.ClientSession
    concurrent_requests_semaphore: asyncio.BoundedSemaphore


async_run_context_var = contextvars.ContextVar[AsyncRunContext]("async_run_context")


def async_run_context() -> AsyncRunContext:
    return async_run_context_var.get()
